#include <gtk/gtk.h>
#include "specimen.h"
#include "lfo.h"

/* widgets */
static GtkWidget *window;
static GtkWidget *lfo_spin;
static GtkWidget *freq_spin;
static GtkWidget *shape_opt;
static GtkWidget *sync_opt;
static GtkWidget *sync_check;

/* adjustments */
static GtkAdjustment *lfo_adj;
static GtkAdjustment *freq_adj;

/* different possible beat lengths */
enum
{
     EIGHTH,
     QUARTER,
     HALF,
     WHOLE,
     DOUBLE,
};

static void cb_close ( )
{
     gtk_widget_hide (window);
}

static void cb_lfo (GtkWidget * widget, gpointer data)
{
     int id;
     float freq;
     float sync;
     gboolean sync_enable;
     LFOShape shape;

     id = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (lfo_spin));

     if (lfo_get_freq (id - 1, &freq) < 0)
     {
	  debug ("%s: failed to get frequency\n", __FUNCTION__);
	  return;
     }
     if (lfo_get_shape (id - 1, &shape) < 0)
     {
	  debug ("%s: failed to get shape\n", __FUNCTION__);
	  return;
     }

     if (lfo_get_sync (id - 1, &sync_enable) < 0)
     {
	  debug ("%s: failed to get sync\n", __FUNCTION__);
	  return;
     }

     if (sync_enable == TRUE)
     {
	  if (lfo_get_sync_beats (id - 1, &sync) < 0)
	  {
	       debug ("%s: failed to get sync beats\n", __FUNCTION__);
	       return;
	  }
     }

     gtk_spin_button_set_value (GTK_SPIN_BUTTON (freq_spin),
				(gdouble) freq);

     switch (shape)
     {
     case LFO_SHAPE_TRIANGLE:
	  gtk_option_menu_set_history (GTK_OPTION_MENU (shape_opt), 0);
	  break;
     case LFO_SHAPE_SINE:
	  gtk_option_menu_set_history (GTK_OPTION_MENU (shape_opt), 1);
	  break;
     case LFO_SHAPE_SAW:
	  gtk_option_menu_set_history (GTK_OPTION_MENU (shape_opt), 2);
	  break;
     case LFO_SHAPE_SQUARE:
	  gtk_option_menu_set_history (GTK_OPTION_MENU (shape_opt), 3);
	  break;
     }

     /* cb_sync will get called if necessary and (de)sensitize appropriately */
     gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (sync_check),
				   sync_enable);

     /* we only need to update the menu if sync_enable is true */
     if (sync_enable)
     {
	  if (sync == 0.5)
	  {
	       gtk_option_menu_set_history (GTK_OPTION_MENU (sync_opt), 0);

	  }
	  else if (sync == 1.0)
	  {
	       gtk_option_menu_set_history (GTK_OPTION_MENU (sync_opt), 1);

	  }
	  else if (sync == 2.0)
	  {
	       gtk_option_menu_set_history (GTK_OPTION_MENU (sync_opt), 2);

	  }
	  else if (sync == 4.0)
	  {
	       gtk_option_menu_set_history (GTK_OPTION_MENU (sync_opt), 3);

	  }
	  else
	  {
	       gtk_option_menu_set_history (GTK_OPTION_MENU (sync_opt), 4);
	  }
     }
}

static void cb_freq (GtkWidget * widget, gpointer data)
{
     int id;
     float freq;

     id = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (lfo_spin));
     freq = gtk_spin_button_get_value (GTK_SPIN_BUTTON (freq_spin));

     if (lfo_set_freq (id - 1, freq) < 0)
	  debug ("%s: failed to set freq\n", __FUNCTION__);
}

static void cb_shape (GtkWidget * widget, LFOShape shape)
{
     int id;

     id = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (lfo_spin));

     if (lfo_set_shape (id - 1, shape) < 0)
	  debug ("%s: failed to set shape\n", __FUNCTION__);
}

static void cb_sync (GtkWidget * widget, gpointer data)
{
     int id;

     id = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (lfo_spin));

     if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget)))
     {

	  gtk_widget_set_sensitive (freq_spin, FALSE);
	  gtk_widget_set_sensitive (sync_opt, TRUE);

	  if (lfo_set_sync (id - 1, TRUE) < 0)
	       debug ("%s: failed to set sync\n", __FUNCTION__);

     }
     else
     {
	  gtk_widget_set_sensitive (freq_spin, TRUE);
	  gtk_widget_set_sensitive (sync_opt, FALSE);

	  if (lfo_set_sync (id - 1, FALSE) < 0)
	       debug ("%s: failed to set sync\n", __FUNCTION__);
     }
}

static void cb_sync_beats (GtkWidget * widget, int beats)
{
     int id;
     int err = 0;

     id = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON (lfo_spin));

     switch (beats)
     {
     case EIGHTH:
	  err = lfo_set_sync_beats (id - 1, 0.5);
	  break;
     case QUARTER:
	  err = lfo_set_sync_beats (id - 1, 1.0);
	  break;
     case HALF:
	  err = lfo_set_sync_beats (id - 1, 2.0);
	  break;
     case WHOLE:
	  err = lfo_set_sync_beats (id - 1, 4.0);
	  break;
     case DOUBLE:
	  err = lfo_set_sync_beats (id - 1, 8.0);
	  break;
     default:
	  err = -1;
     }

     if (err < 0)
	  debug ("%s: failed to set sync beats\n", __FUNCTION__);
}

void lfo_settings_show ( )
{
     gtk_widget_show (window);
}

void lfo_settings_init (GtkWidget * parent)
{
     GtkWidget *vbox;
     GtkWidget *hbox;
     GtkWidget *label;
     GtkWidget *menu;
     GtkWidget *item;
     GtkWidget *button;
     GtkWidget *sep;

     debug ("Initializing LFO settings window\n");

     /* main window */
     window = gtk_window_new (GTK_WINDOW_TOPLEVEL);
     gtk_window_set_title (GTK_WINDOW (window), "LFOs");
     gtk_window_set_resizable (GTK_WINDOW (window), FALSE);
     gtk_window_set_transient_for (GTK_WINDOW (window),
				   GTK_WINDOW (parent));
     gtk_window_set_modal (GTK_WINDOW (window), TRUE);
     g_signal_connect (GTK_WINDOW (window), "delete-event",
		       G_CALLBACK (cb_close), NULL);

     /* vbox */
     vbox = gtk_vbox_new (FALSE, SPACING);
     gtk_container_add (GTK_CONTAINER (window), vbox);
     gtk_container_set_border_width (GTK_CONTAINER (window), SPACING);
     gtk_widget_show (vbox);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /*******/
     /* LFO */
     /*******/
     label = gtk_label_new ("LFO:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     lfo_adj =
	  (GtkAdjustment *) gtk_adjustment_new (1.0, 1.0, LFO_COUNT, 1.0,
						0.0, 0.0);

     /* spin */
     lfo_spin = gtk_spin_button_new (lfo_adj, 0, 0);
     gtk_box_pack_end (GTK_BOX (hbox), lfo_spin, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (lfo_adj), "value-changed",
		       G_CALLBACK (cb_lfo), NULL);
     gtk_widget_show (lfo_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /********/
     /* Freq */
     /********/
     label = gtk_label_new ("Freq:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* adjustment */
     freq_adj =
	  (GtkAdjustment *) gtk_adjustment_new (1.0, 0.0, 20.0, 0.01, 0.1,
						1.0);

     /* spin */
     freq_spin = gtk_spin_button_new (freq_adj, 0.1, 2);
     gtk_box_pack_end (GTK_BOX (hbox), freq_spin, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (freq_adj), "value-changed",
		       G_CALLBACK (cb_freq), NULL);
     gtk_widget_show (freq_spin);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /*********/
     /* Sync  */
     /*********/
     label = gtk_label_new ("Sync:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* sync check box */
     sync_check = gtk_check_button_new ( );
     gtk_box_pack_start (GTK_BOX (hbox), sync_check, FALSE, FALSE, 0);
     g_signal_connect (G_OBJECT (sync_check), "toggled",
		       G_CALLBACK (cb_sync), (gpointer) sync_check);
     gtk_widget_show (sync_check);

     /* sync menu */
     menu = gtk_menu_new ( );

     /* the different note size, the values are in beats */
     item = gtk_menu_item_new_with_label ("0.5 beats");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_sync_beats), (gpointer) EIGHTH);
     gtk_widget_show (item);

     item = gtk_menu_item_new_with_label ("1 beat");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_sync_beats), (gpointer) QUARTER);
     gtk_widget_show (item);

     item = gtk_menu_item_new_with_label ("2 beats");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_sync_beats), (gpointer) HALF);
     gtk_widget_show (item);

     item = gtk_menu_item_new_with_label ("4 beats");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_sync_beats), (gpointer) WHOLE);
     gtk_widget_show (item);

     item = gtk_menu_item_new_with_label ("8 beats");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate",
		       G_CALLBACK (cb_sync_beats), (gpointer) DOUBLE);
     gtk_widget_show (item);

     /* sync optionmenu */
     sync_opt = gtk_option_menu_new ( );
     gtk_box_pack_end (GTK_BOX (hbox), sync_opt, FALSE, FALSE, 0);
     gtk_option_menu_set_menu (GTK_OPTION_MENU (sync_opt), menu);
     gtk_widget_show (sync_opt);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /*********/
     /* Shape */
     /*********/
     label = gtk_label_new ("Shape:");
     gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 0);
     gtk_widget_show (label);

     /* shape menu */
     menu = gtk_menu_new ( );

     /* triangle */
     item = gtk_menu_item_new_with_label ("Triangle");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_shape),
		       (gpointer) LFO_SHAPE_TRIANGLE);
     gtk_widget_show (item);

     /* sine */
     item = gtk_menu_item_new_with_label ("Sine");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_shape),
		       (gpointer) LFO_SHAPE_SINE);
     gtk_widget_show (item);

     /* saw */
     item = gtk_menu_item_new_with_label ("Saw");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_shape),
		       (gpointer) LFO_SHAPE_SAW);
     gtk_widget_show (item);

     /* square */
     item = gtk_menu_item_new_with_label ("Square");
     gtk_menu_shell_append (GTK_MENU_SHELL (menu), item);
     g_signal_connect (G_OBJECT (item), "activate", G_CALLBACK (cb_shape),
		       (gpointer) LFO_SHAPE_SQUARE);
     gtk_widget_show (item);

     /* shape optionmenu */
     shape_opt = gtk_option_menu_new ( );
     gtk_box_pack_end (GTK_BOX (hbox), shape_opt, FALSE, FALSE, 0);
     gtk_option_menu_set_menu (GTK_OPTION_MENU (shape_opt), menu);
     gtk_widget_show (shape_opt);

     /**************/
     sep = gtk_hseparator_new ( );
     gtk_box_pack_start (GTK_BOX (vbox), sep, FALSE, FALSE, 0);
     gtk_widget_show (sep);

     /* hbox */
     hbox = gtk_hbox_new (FALSE, SPACING);
     gtk_box_pack_start (GTK_BOX (vbox), hbox, FALSE, FALSE, 0);
     gtk_widget_show (hbox);

     /* close button */
     button = gtk_button_new_from_stock (GTK_STOCK_CLOSE);
     g_signal_connect (G_OBJECT (button), "clicked", G_CALLBACK (cb_close),
		       NULL);
     gtk_box_pack_end (GTK_BOX (hbox), button, FALSE, FALSE, 0);
     gtk_widget_show (button);

     /* initialize display (theoretically fetching the settings of the
      * first LFO) */
     g_signal_emit_by_name (G_OBJECT (lfo_adj), "value-changed");

     /* this is necessary the first time through to ensure that if the default patch configuration
      * is without midi syncing the beat menu will get grayed out */
     g_signal_emit_by_name (G_OBJECT (sync_check), "toggled");
}
