#include <math.h>
#include <glib.h>
#include "maths.h"
#include "specimen.h"
#include "lfo.h"
#include "driver.h"

typedef struct _LFO
{
     float val;			/* current value */
     guint32 phase;		/* current phase within shape, 8 MSB
				 * representing integer part and 24
				 * LSB representing fractional part */
     float freq;			/* frequency in Hz */
     guint32 inc;		/* amount to increase phase by per tick */
     LFOShape shape;		/* the shape we are using */
     float table[257];		/* waveform table (size is 257 so that
				 * we can do table[n+1] even if n is
				 * 255, which is handy in lerping
				 * code) */
     gboolean sync;		/* whether we use tempo syncing or not */
     float sync_beats;		/* if we sync, how many beats we sync to */
}
LFO;

/* LFOs at our disposal */
static LFO lfos[LFO_COUNT];

/* sample rate we expect the world to be running at */
static int samplerate = DRIVER_DEFAULT_SAMPLERATE;

/* make sure all LFOs are in reasonable shape */
void lfo_init ( )
{
     int i;

     /* set each LFO to "sensible" defaults */
     for (i = 0; i < LFO_COUNT; i++)
     {
	  lfos[i].val = 0.0;
	  lfos[i].phase = 0;
	  lfos[i].sync = FALSE;
	  lfos[i].sync_beats = 1.0;
	  lfo_set_freq (i, 1.0);
	  lfo_set_shape (i, LFO_SHAPE_TRIANGLE);
     }
}

/* set the sample rate we expect the world to be running at; we don't
 * have to do any locking here because this function is guaranteed to
 * be called only when we are not processing (and processing will wait
 * until we are finished) */
void lfo_set_samplerate (int rate)
{
     int i;

     /* reset the phase and inc values of each LFO since they are
      * dependent on the samplerate */
     if (rate != samplerate)
	  for (i = 0; i < LFO_COUNT; i++)
	       lfo_set_freq (i, lfos[i].freq);

     samplerate = rate;
}

/* set the frequency this LFO oscillates at */
int lfo_set_freq (int id, float freq)
{
     if (id < 0 || id >= LFO_COUNT)
	  return LFO_ERR_ID;

     if (freq < 0)
	  return LFO_ERR_PARAM;

     lfos[id].inc =
	  (guint) ((256.0 * freq / samplerate) * (float) (1 << 24));
     lfos[id].freq = freq;

     return 0;
}

/* build a table for our LFO shape */
int lfo_set_shape (int id, LFOShape shape)
{
     int i;
     float *t;

     if (id < 0 || id >= LFO_COUNT)
	  return LFO_ERR_ID;

     t = lfos[id].table;

     switch (shape)
     {
     case LFO_SHAPE_TRIANGLE:
	  for (i = 0; i < 64; i++)
	  {
	       t[i] = i / 64.0;
	       t[i + 64] = (64 - i) / 64.0;
	       t[i + 128] = -i / 64.0;
	       t[i + 192] = -(64 - i) / 64.0;
	  }
	  t[256] = 0.0;
	  break;
     case LFO_SHAPE_SINE:
	  for (i = 0; i <= 256; i++)
	       t[i] = sin (2.0 * G_PI * (i / 256.0));
	  break;
     case LFO_SHAPE_SAW:
	  for (i = 0; i < 256; i++)
	       t[i] = 2.0 * (i / 255.0) - 1.0;
	  t[256] = -1.0;
	  break;
     case LFO_SHAPE_SQUARE:
	  for (i = 0; i < 128; i++)
	  {
	       t[i] = 1.0;
	       t[i + 128] = -1.0;
	  }
	  t[256] = 1.0;
	  break;
     default:
	  return LFO_ERR_PARAM;
	  break;
     }

     lfos[id].shape = shape;
     return 0;
}

/* set whether this LFO is tempo syncable or not */
int lfo_set_sync (int id, gboolean state)
{
     if (id < 0 || id >= LFO_COUNT)
	  return LFO_ERR_ID;

     lfos[id].sync = state;
     return 0;
}

/* set the period length of this LFO if it's syncable */
int lfo_set_sync_beats (int id, float beats)
{
     if (id < 0 || id >= LFO_COUNT)
	  return LFO_ERR_ID;

     if (!lfos[id].sync)
	  return LFO_ERR_ID;

     lfos[id].sync_beats = beats;
     return 0;
}

/* get the frequency we run at */
int lfo_get_freq (int id, float *val)
{
     if (id < 0 || id >= LFO_COUNT)
	  return LFO_ERR_ID;

     *val = lfos[id].freq;
     return 0;
}

/* get the shape we use */
int lfo_get_shape (int id, LFOShape * val)
{
     if (id < 0 || id >= LFO_COUNT)
	  return LFO_ERR_ID;

     *val = lfos[id].shape;
     return 0;
}

/* get whether this LFO syncs or not */
int lfo_get_sync (int id, gboolean * val)
{
     if (id < 0 || id >= LFO_COUNT)
	  return LFO_ERR_ID;

     *val = lfos[id].sync;
     return 0;
}

/* get the number of beats this LFO syncs to */
int lfo_get_sync_beats (int id, float *val)
{
     if (id < 0 || id >= LFO_COUNT)
	  return LFO_ERR_ID;

     if (!lfos[id].sync)
	  return LFO_ERR_ID;

     *val = lfos[id].sync_beats;
     return 0;
}

/* start all syncable LFOs */
int lfo_sync (float bpm)
{
     int i;
     float freq;

     if (bpm < 0.0)
	  return LFO_ERR_PARAM;

     for (i = 0; i < LFO_COUNT; i++)
     {
	  if (lfos[i].sync)
	  {

	       /* calculate new frequency */
	       freq = (bpm / 60.0) / (lfos[i].sync_beats);
	       lfo_set_freq (i, freq);

	       /* restart */
	       lfos[i].phase = 0;
	  }
     }

     return 0;
}

/* advance all LFO's phase by nframes */
void lfo_advance_all (int nframes)
{
     int i;
     guint8 index;
//     float frac;
     guint8 frac;
     LFO *lfop;

     for (i = 0; i < LFO_COUNT; i++)
     {
	  lfop = &lfos[i];

	  lfop->phase += lfop->inc * nframes;

	  /* calculate new value */
	  index = lfop->phase >> 24;
//	  frac = (lfop->phase & 0x00FFFFFF) / (1 << 24);
	  frac = (lfop->phase & 0x00FFFFFF) >> 16;
	  
	  /* linear interpolation */
/*	  lfop->val = lfop->table[index] * (1.0 - frac)
	  + lfop->table[index + 1] * frac; */
	  lfop->val = cerp (lfop->table[(index == 0)? 255: index],
			    lfop->table[index],
			    lfop->table[index+1],
			    lfop->table[index+2],
			    frac);
     }
}

/* get the current value of the lfo [-1.0, 1.0] */
float lfo_get_val (int id)
{
     if (id < 0 || id >= LFO_COUNT)
	  return 0;

     return lfos[id].val;
}
