#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <pthread.h>
#include <glib.h>
#include "specimen.h"
#include "maths.h"
#include "ticks.h"
#include "patch.h"
#include "sample.h"
#include "adsr.h"
#include "lfo.h"
#include "driver.h"		/* for DRIVER_DEFAULT_SAMPLERATE */

/* the minimum envelope release value we'll allow (to prevent clicks) */
const float PATCH_MIN_RELEASE = 0.05;

/* what sample rate we think the audio interface is running at */
static float samplerate = DRIVER_DEFAULT_SAMPLERATE;

/* type for currently playing notes (voices) */
typedef struct _PatchVoice
{
     gboolean active;		/* whether this voices is playing or not */
     Tick ticks;		/* at what time this voice was activated */
     int offset;		/* how many ticks should pass before this voice is used */
     int relset;		/* how many ticks should pass before we release this voice (negative if N/A) */
     gboolean released;		/* whether or not we've been released or not */
     int dir;			/* current direction (forward == 1, reverse == -1) */
     int note;			/* the note that activated us */
     int posi;			/* integer sample index */
     guint32 posf;		/* fractional sample index */
     int stepi;			/* integer step amount */
     guint32 stepf;		/* fractional step amount */
     float vel;			/* velocity; volume of this voice */
     ADSR vol;			/* volume envelope */
     ADSR pan;			/* panning envelope */
     ADSR ffreq;		/* filter frequency envelope */
     ADSR freso;		/* filter resonance envelope */
     float fll;			/* lowpass filter buffer, left */
     float flr;			/* lowpass filter buffer, right */
     float fbl;			/* bandpass filter buffer */
     float fbr;			/* bandpass filter buffer, right */
}
PatchVoice;

/* type for modulatable parameters */
typedef struct _PatchParam
{
     float val;			/* value of this parameter */
     ADSRParams env;		/* envelope parameters */
     float env_amt;		/* amount of envelope we add [-1.0, 1.0] */
     int lfo;			/* which lfo we use, or negative if none */
     float lfo_amt;		/* amount of lfo we add [-1.0, 1.0] */
}
PatchParam;

/* type for array of instruments (called patches) */
typedef struct _Patch
{
     int active;		/* whether this instance of patch_s is in use or not */
     Sample sample;		/* sample data */
     int display_index;		/* order in which it is to be displayed */
     char name[PATCH_MAX_NAME];	/* string name */
     int channel;		/* midi channel to listen on */
     int note;			/* midi note to listen on */
     int range;			/* whether to listen to a range of notes or not */
     int lower_note;		/* lowest note in range */
     int upper_note;		/* highest note in range */
     int cut;			/* cut signal this patch emits */
     int cut_by;		/* what cut signals stop this patch */
     int sample_start;		/* the first frame to play */
     int sample_stop;		/* the last frame to play */
     int loop_start;		/* the first frame to loop at */
     int loop_stop;		/* the last frame to loop at */

     PatchPlayMode play_mode;	/* how this patch is to be played */
     PatchParam vol;		/* volume:                  [0.0, 1.0] */
     PatchParam pan;		/* panning:                [-1.0, 1.0] */
     PatchParam ffreq;	        /* filter cutoff frequency: [0.0, 1.0] */
     PatchParam freso;	        /* filter resonance:        [0.0, 1.0] */

     /* each patch is responsible for its own voices */
     PatchVoice voices[PATCH_VOICE_COUNT];
}
Patch;

/* array of all patches  */
static Patch patches[PATCH_COUNT];

/*****************************************************************************/
/************************ PRIVATE GENERAL HELPER FUNCTIONS********************/
/*****************************************************************************/

/* verifies that a given id refers to a valid patch */
inline static int isok (int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patches[id].active)
	  return 0;

     return 1;
}

/* a utility function which finds the PatchParam* associated with
 * a particular PatchParamType */
static int patch_get_param (PatchParam ** p, int id, PatchParamType param)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     switch (param)
     {
     case PATCH_PARAM_VOLUME:
	  *p = &patches[id].vol;
	  break;
     case PATCH_PARAM_PANNING:
	  *p = &patches[id].pan;
	  break;
     case PATCH_PARAM_CUTOFF:
	  *p = &patches[id].ffreq;
	  break;
     case PATCH_PARAM_RESONANCE:
	  *p = &patches[id].freso;
	  break;
     default:
	  return PATCH_PARAM_INVALID;
	  break;
     }

     return 0;
}

/*****************************************************************************/
/************************ UTILITY FUNCTIONS **********************************/
/*****************************************************************************/

/* returns the number of patches currently active */
int patch_count ( )
{
     int id, count;

     for (id = count = 0; id < PATCH_COUNT; id++)
	  if (patches[id].active)
	       count++;

     return count;
}

/* returns assigned patch id on success, negative value on failure */
int patch_create (const char *name)
{
     int id, i;
     ADSRParams defadsr;
     PatchVoice defvoice;

     /* find an unoccupied patch id */
     for (id = 0; patches[id].active; id++)
	  if (id == PATCH_COUNT)
	       return PATCH_LIMIT;

     patches[id].active = 1;

     debug ("Creating patch %s (%d).\n", name, id);

     /* name */
     strncpy (patches[id].name, name, PATCH_MAX_NAME);
     patches[id].name[PATCH_MAX_NAME - 1] = '\0';
     patches[id].sample.name[0] = '\0';

     /* sample */
     pthread_mutex_lock (&patches[id].sample.mutex);
     patches[id].sample.sp = NULL;
     pthread_mutex_unlock (&patches[id].sample.mutex);

     /* default values */
     patches[id].sample.frames = 0;
     patches[id].channel = 0;
     patches[id].note = 60;
     patches[id].range = 0;
     patches[id].lower_note = 60;
     patches[id].upper_note = 60;
     patches[id].play_mode = PATCH_PLAY_FORWARD | PATCH_PLAY_SINGLESHOT;
     patches[id].cut = 0;
     patches[id].cut_by = 0;
     patches[id].sample_start = 0;
     patches[id].sample_stop = 0;
     patches[id].loop_start = 0;
     patches[id].loop_stop = 0;

     /* default adsr params */
     defadsr.attack = 0.0;
     defadsr.decay = 0.0;
     defadsr.sustain = 1.0;
     defadsr.release = 0.0;

     /* volume */
     patches[id].vol.val = 1.0;
     patches[id].vol.env = defadsr;
     patches[id].vol.env_amt = 0.0;
     patches[id].vol.lfo = -1;
     patches[id].vol.lfo_amt = 0.0;

     /* panning */
     patches[id].pan.val = 0.0;
     patches[id].pan.env = defadsr;
     patches[id].pan.env_amt = 0.0;
     patches[id].pan.lfo = -1;
     patches[id].pan.lfo_amt = 0.0;

     /* cutoff */
     patches[id].ffreq.val = 1.0;
     patches[id].ffreq.env = defadsr;
     patches[id].ffreq.env_amt = 0.0;
     patches[id].ffreq.lfo = -1;
     patches[id].ffreq.lfo_amt = 0.0;

     /* resonance */
     patches[id].freso.val = 0.0;
     patches[id].freso.env = defadsr;
     patches[id].freso.env_amt = 0.0;
     patches[id].freso.lfo = -1;
     patches[id].freso.lfo_amt = 0.0;

     /* default voice */
     defvoice.active = FALSE;
     defvoice.offset = 0;
     defvoice.note = 0;
     defvoice.posi = 0;
     defvoice.posf = 0;
     defvoice.stepi = 0;
     defvoice.stepf = 0;
     defvoice.vel = 0;
     adsr_init (&defvoice.vol);
     adsr_init (&defvoice.pan);
     adsr_init (&defvoice.ffreq);
     adsr_init (&defvoice.freso);

     /* initialize voices */
     for (i = 0; i < PATCH_VOICE_COUNT; i++)
	  patches[id].voices[i] = defvoice;

     /* set display_index to next unique value */
     patches[id].display_index = 0;
     for (i = 0; i < PATCH_COUNT; i++)
     {
	  if (i == id)
	       continue;
	  if (patches[i].active
	      && patches[i].display_index >= patches[id].display_index)
	  {
	       patches[id].display_index = patches[i].display_index + 1;
	  }
     }

     return id;
}

/* destroy a single patch with given id */
int patch_destroy (int id)
{
     int index;

     if (!isok (id))
	  return PATCH_ID_INVALID;
     debug ("Removing patch: %d\n", id);

     patches[id].active = 0;
     pthread_mutex_lock (&patches[id].sample.mutex);
     if (patches[id].sample.sp != NULL)
     {
	  free (patches[id].sample.sp);
	  patches[id].sample.sp = NULL;
     }
     pthread_mutex_unlock (&patches[id].sample.mutex);

     /* every active patch with a display_index greater than
      * this patch's needs to have it's value decremented so
      * that we preservere continuity */
     index = patches[id].display_index;
     for (id = 0; id < PATCH_COUNT; id++)
     {
	  if (patches[id].active && patches[id].display_index > index)
	       patches[id].display_index--;
     }

     return 0;
}

/* destroy all patches */
void patch_destroy_all ( )
{
     int id;

     for (id = 0; id < PATCH_COUNT; id++)
	  patch_destroy (id);

     return;
}

/* place all patch ids, sorted in ascending order by display index,
   into array 'id' and return number of patches */
int patch_dump (int **dump)
{
     int i, j, id, count, tmp;

     *dump = NULL;

     /* determine number of patches */
     count = patch_count ( );

     if (count == 0)
	  return count;

     /* allocate dump */
     *dump = malloc (sizeof (int) * count);
     if (*dump == NULL)
	  return PATCH_ALLOC_FAIL;

     /* place active patches into dump array */
     for (id = i = 0; id < PATCH_COUNT; id++)
	  if (patches[id].active)
	       (*dump)[i++] = id;

     /* sort dump array by display_index in ascending order */
     for (i = 0; i < count; i++)
     {
	  for (j = i; j < count; j++)
	  {
	       if (patches[(*dump)[j]].display_index <
		   patches[(*dump)[i]].display_index)
	       {
		    tmp = (*dump)[i];
		    (*dump)[i] = (*dump)[j];
		    (*dump)[j] = tmp;
	       }
	  }
     }

     return count;
}

int patch_duplicate (int target)
{
     int id, i;

     if (target < 0 || target > PATCH_COUNT || !patches[target].active)
	  return PATCH_ID_INVALID;

     /* find an unoccupied patch id */
     for (id = 0; patches[id].active; id++)
	  if (id == PATCH_COUNT)
	       return PATCH_LIMIT;

     debug ("Creating patch (%d) from patch %s (%d).\n", id,
	    patches[target].name, target);

     patches[id] = patches[target];

     /* we don't want to share memory with the target patch */
     pthread_mutex_lock (&patches[id].sample.mutex);
     patches[id].sample.sp = NULL;
     pthread_mutex_unlock (&patches[id].sample.mutex);

     /* load same soundfile as target patch */
     if (patches[target].sample.sp != NULL)
     {
	  patch_sample_load (id, patches[target].sample.name);
     }

     /* set display_index to next unique value */
     patches[id].display_index = 0;
     for (i = 0; i < PATCH_COUNT; i++)
     {
	  if (i == id)
	       continue;
	  if (patches[i].active
	      && patches[i].display_index >= patches[id].display_index)
	  {
	       patches[id].display_index = patches[i].display_index + 1;
	  }
     }

     debug ("chosen display: %d\n", patches[id].display_index);
     return id;
}

/* stop all currently playing voices in given patch */
int patch_flush (int id)
{
     int i;
     
     if (!isok(id))
	  return PATCH_ID_INVALID;

     pthread_mutex_lock (&patches[id].sample.mutex);
     if (patches[id].sample.sp == NULL)
     {
	  pthread_mutex_unlock (&patches[id].sample.mutex);
	  return 0;
     }

     for (i = 0; i < PATCH_VOICE_COUNT; i++)
	  patches[id].voices[i].active = FALSE;
	  
     pthread_mutex_unlock (&patches[id].sample.mutex);
     return 0;
}

/* stop all voices for all patches */
void patch_flush_all ( )
{
     int i;

     for (i = 0; i < PATCH_COUNT; i++)
     {
	  patch_flush (i);
     }
}    
	  
/* returns error message associated with error code */
const char *patch_strerror (int error)
{
     switch (error)
     {
     case PATCH_PARAM_INVALID:
	  return "patch parameter is invalid";
	  break;
     case PATCH_ID_INVALID:
	  return "patch id is invalid";
	  break;
     case PATCH_ALLOC_FAIL:
	  return "failed to allocate space for patch";
	  break;
     case PATCH_NOTE_INVALID:
	  return "specified note is invalid";
	  break;
     case PATCH_PAN_INVALID:
	  return "specified panning is invalid";
	  break;
     case PATCH_CHANNEL_INVALID:
	  return "specified channel is invalid";
	  break;
     case PATCH_VOL_INVALID:
	  return "specified volume is invalid";
	  break;
     case PATCH_MASTER_VOLUME_INVALID:
	  return "specified master volume is invalid";
	  break;
     case PATCH_PLAY_MODE_INVALID:
	  return "specified patch play mode is invalid";
	  break;
     case PATCH_LIMIT:
	  return "maximum patch count reached, can't create another";
	  break;
     case PATCH_SAMPLE_INDEX_INVALID:
	  return "specified sample is invalid";
	  break;
     default:
	  return "unknown error";
     }
}

/* loads a sample file for a patch */
int patch_sample_load (int id, char *name)
{
     int val;

     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (name == NULL)
     {
	  debug ("Refusing to load null sample for patch %d\n", id);
	  return PATCH_PARAM_INVALID;
     }

     debug ("Loading sample %s for patch %d\n", name, id);
     patch_flush (id);
     val = sample_load_file (&patches[id].sample, name, samplerate);

     /* set the sample/loop start/stop point appropriately */
     patches[id].sample_start = 0;
     patches[id].sample_stop = patches[id].sample.frames - 1;
     patches[id].loop_start = 0;
     patches[id].loop_stop = patches[id].sample.frames - 1;

     return val;
}

/* sets our samplerate and resamples if necessary */
void patch_set_samplerate (int rate)
{
     int id;
     char *name;
     int oldrate = samplerate;

     samplerate = rate;
     
     if (samplerate != oldrate)
     {	 
	  for (id = 0; id < PATCH_COUNT; id++)
	  {
	       if (!patches[id].active)
		    continue;
	       name = patch_get_sample_name (id);
	       patch_sample_load (id, name);
	       if (name != NULL)
		    free (name);
	  }
     }
}

/* unloads a patch's sample */
void patch_sample_unload (int id)
{
     debug ("Unloading sample for patch %d\n", id);
     pthread_mutex_lock (&patches[id].sample.mutex);

     if (patches[id].sample.sp != NULL)
     {
	  free (patches[id].sample.sp);
	  patches[id].sample.sp = NULL;
     }

     patches[id].sample.frames = 0;
     patches[id].sample_start = 0;
     patches[id].sample_stop = 0;
     patches[id].loop_start = 0;
     patches[id].loop_stop = 0;
     patches[id].sample.name[0] = '\0';

     pthread_mutex_unlock (&patches[id].sample.mutex);
}

/*****************************************************************************/
/******************* PLAYBACK AND RENDERING FUNCTIONS ************************/
/*****************************************************************************/

/* a helper function to release all voices matching a given criteria
 * (if note is a negative value, all active voices will be released) */
inline static void patch_release_patch (Patch * p, int note, int offset)
{
     int i;
     PatchVoice *v;

     for (i = 0; i < PATCH_VOICE_COUNT; i++)
     {
	  if (p->voices[i].active && (p->voices[i].note == note || note < 0))
	  {
	       v = &p->voices[i];

	       /* we don't really release here, that's the job of
		* advance( ); we just tell it *when* to release */
	       v->relset = offset;
	  }
     }
}

/* a helper function to cut all patches whose cut_by value matches the
 * cut value for the given patch */
inline static void patch_cut_patch (Patch * p, int offset)
{
     int i;

     /* a cut value of zero is ignored so that the user has a way of *not* using
      * cuts*/
     if (p->cut == 0)
	  return;

     for (i = 0; i < PATCH_COUNT; i++)
     {
	  if (patches[i].active && patches[i].cut_by == p->cut)
	  {
	       patch_release_patch (&patches[i], -69, offset);
	  }
     }
}

/* a helper function to activate a voice for a patch */
inline static void patch_trigger_patch (Patch * p, int note, float vel,
					Tick ticks, int offset)
{
     register int i;
     PatchVoice *v;
     double tmp;
     int gzr = 0;		/* index of the oldest running patch */
     Tick oldest = ticks;	/* time of the oldest running patch */

     /* we don't activate voices if we don't have a sample */
     if (p->sample.sp == NULL)
	  return;

     /* find a free voice slot and determine the oldest running voice */
     for (i = 0; i < PATCH_VOICE_COUNT; i++)
     {
	  if (p->voices[i].ticks <= oldest)
	  {
	       oldest = p->voices[i].ticks;
	       gzr = i;
	  }
	  if (!p->voices[i].active)
	  {
	       break;
	  }
     }

     /* take the oldest running voice's slot if we couldn't find an
      * empty one */
     if (i == PATCH_VOICE_COUNT)
     {
	  v = &p->voices[gzr];
     }
     else
     {
	  v = &p->voices[i];
     }

     /* fill in our voice */
     v->ticks = ticks;
     v->offset = offset;
     v->relset = -1;		/* N/A at this time */
     v->released = FALSE;
     v->note = note;
     v->fll = 0;
     v->fbl = 0;
     v->flr = 0;
     v->fbr = 0;

     /* setup direction */
     if (p->play_mode & PATCH_PLAY_REVERSE)
     {
	  v->posi = p->sample_stop;
	  v->posf = 0;
	  v->dir = -1;
     }
     else
     {
	  v->posi = p->sample_start;
	  v->posf = 0;
	  v->dir = 1;
     }

     /* setup pitch scaling */
     tmp = pow (2, (note - p->note) / 12.0);
     v->stepi = tmp;
     v->stepf = (tmp - v->stepi) * (0xFFFFFFFFU); /* maximum value of guint32 */
     v->vel = vel;

     /* setup envelopes */
     adsr_set_params (&v->vol, &p->vol.env);
     adsr_set_params (&v->pan, &p->pan.env);
     adsr_set_params (&v->ffreq, &p->ffreq.env);
     adsr_set_params (&v->freso, &p->freso.env);

     /* trigger envelopes */
     adsr_trigger (&v->vol);
     adsr_trigger (&v->pan);
     adsr_trigger (&v->ffreq);
     adsr_trigger (&v->freso);

     /* mark our territory */
     v->active = TRUE;
}

/* a helper routine to determine the pitch-scaled sample values to use for a frame */
inline static void pitchscale (Patch * p, PatchVoice * v, float *l,
			       float *r)
{
     int y0, y1, y2, y3;

     /* determine sample indices */
     if ((y0 = (v->posi - 1) * 2) < 0)
	  y0 = 0;
     y1 = v->posi * 2;
     if ((y2 = (v->posi + 1) * 2) >= p->sample.frames * 2)
	  y2 = 0;
     if ((y3 = (v->posi + 2) * 2) >= p->sample.frames * 2)
	  y3 = 0;

     /* interpolate */
     *l = cerp (p->sample.sp[y0],
		p->sample.sp[y1],
		p->sample.sp[y2], p->sample.sp[y3], v->posf >> 24);
     *r = cerp (p->sample.sp[y0 + 1],
		p->sample.sp[y1 + 1],
		p->sample.sp[y2 + 1], p->sample.sp[y3 + 1], v->posf >> 24);
}

/* a helper routine to setup panning of two samples in a frame  */
inline static void pan (Patch * p, PatchVoice * v, float *l, float *r)
{
     float pan = p->pan.val + (adsr_tick (&v->pan) * p->pan.env_amt);

     if (p->pan.lfo >= 0)
	  pan += lfo_get_val (p->pan.lfo) * p->pan.lfo_amt;

     /* determine values after modulation */
     if (pan > 1.0)
	  pan = 1.0;
     else if (pan < -1.0)
	  pan = -1.0;

     if (pan < 0.0)
     {				/* panned left */
	  *l += *r * -pan;
	  *r *= 1 + pan;
     }
     else if (pan > 0.0)
     {				/* panned right */
	  *r += *l * pan;
	  *l *= 1 - pan;
     }
}

/* a helper routine to apply filters to a frame */
inline static void filter (Patch * p, PatchVoice * v, float *l,
			   float *r)
{
     float ffreq, freso;

     /* determine ffreq value */
     ffreq = p->ffreq.val + (adsr_tick (&v->ffreq) * p->ffreq.env_amt);
     if (p->ffreq.lfo >= 0)
	  ffreq += lfo_get_val (p->ffreq.lfo) * p->ffreq.lfo_amt;

     /* clip */
     if (ffreq > 1.0)
	  ffreq = 1.0;
     else if (ffreq < 0.0)
	  ffreq = 0.0;

     /* determine freso value */
     freso = p->freso.val + (adsr_tick (&v->freso) * p->freso.env_amt);
     if (p->freso.lfo >= 0)
	  freso += lfo_get_val (p->freso.lfo) * p->freso.lfo_amt;

     /* clip */
     if (freso > 1.0)
	  freso = 1.0;
     else if (freso < 0.0)
	  freso = 0.0;

     /* left */
     v->fbl = freso * v->fbl + ffreq * (*l - v->fll);
     v->fll += ffreq * v->fbl;
     *l = v->fll;

     /* right */
     v->fbr = freso * v->fbr + ffreq * (*r - v->flr);
     v->flr += ffreq * v->fbr;
     *r = v->flr;
}

/* a helper routine to adjust the volume of a frame */
inline static int gain (Patch * p, PatchVoice * v, float *l, float *r)
{
     float vol;

     /* determine amount to scale by */
     vol = p->vol.val * adsr_tick (&v->vol);
     if (p->vol.lfo >= 0)
	  vol += lfo_get_val (p->vol.lfo) * p->vol.lfo_amt;

     /* clip */
     if (vol > 1.0)
	  vol = 1.0;
     else if (vol < 0.0)
	  vol = 0.0;

     /* scale */
     *l *= v->vel * vol;
     *r *= v->vel * vol;

     /* check to see if we've finished a release */
     if (vol == 0.0 && v->released)
	  return -1;

     return 0;
}

/* a helper routine to advance to the next frame while properly
 * accounting for the different possible play modes (negative value
 * returned if we are out of samples after doing our work) */
inline static int advance (Patch * p, PatchVoice * v)
{
     guint32 next_posf;

     /* advance our position indices */
     if (v->dir > 0)
     {
	  next_posf = v->posf + v->stepf;
	  if (next_posf < v->posf)	/* unsigned int wraps around */
	       v->posi++;
	  v->posf = next_posf;
	  v->posi += v->stepi;
     }
     else
     {
	  next_posf = v->posf + v->stepf;
	  if (next_posf < v->posf)	/* unsigned int wraps around */
	       v->posi--;
	  v->posf = next_posf;
	  v->posi -= v->stepi;
     }

     /* adjust our indices according to our play mode */
     if (p->play_mode & PATCH_PLAY_LOOP)
     {
	  if (p->play_mode & PATCH_PLAY_PINGPONG)
	  {
	       if ((v->dir > 0) && (v->posi > p->loop_stop))
	       {
		    v->posi = p->loop_stop;
		    v->dir = -1;
	       }
	       else if ((v->dir < 0) && (v->posi < p->loop_start))
	       {
		    v->posi = p->loop_start;
		    v->dir = 1;
	       }
	  }
	  else
	  {
	       if ((v->dir > 0) && (v->posi > p->loop_stop))
	       {
		    v->posi = p->loop_start;
	       }
	       else if ((v->dir < 0) && (v->posi < p->loop_start))
	       {
		    v->posi = p->loop_stop;
	       }
	  }
     }
     else
     {
	  if (((v->dir > 0) && (v->posi > p->sample_stop))
	      || ((v->dir < 0) && (v->posi < p->sample_start)))
	  {

	       /* we need to let our caller know that they are out of
		* samples */
	       return -1;
	  }
     }

     /* check to see if it's time to release */
     if (v->relset >= 0 && !v->released)
     {
	  if (v->relset == 0)
	  {
	       adsr_release (&v->vol);
	       adsr_release (&v->pan);
	       adsr_release (&v->ffreq);
	       adsr_release (&v->freso);
	       v->released = TRUE;
	  }
	  else
	  {
	       v->relset--;
	  }
     }

     return 0;
}

/* a helper rountine to render all active voices of a given patch into buf */
inline static void patch_render_patch (Patch* p, float* buf, int nframes)
{
     register int i;
     register int j;
     PatchVoice* v;
     float l, r;
     gboolean done;

     for (i = 0; i < PATCH_VOICE_COUNT; i++)
     {
	  if (p->voices[i].active == TRUE)
	  {
	       /* sanity check */
	       if (p->voices[i].posi < p->sample.frames)
	       {
		    v = &p->voices[i];
	       }
	       else
	       {
		    p->voices[i].active = 0;
		    continue;
	       }

	       done = FALSE;
	       for (j = v->offset; j < nframes && done == FALSE; j++)
	       {

		    /* process samples */
		    pitchscale (p, v, &l, &r);
		    pan (p, v, &l, &r);
		    filter (p, v, &l, &r);

		    /* adjust volume and stop rendering if we finished
		     * a release */
		    if (gain (p, v, &l, &r) < 0)
			 done = TRUE;

		    buf[j * 2] += l;
		    buf[j * 2 + 1] += r;

		    /* advance our position and stop rendering if we
		     * run out of samples */
		    if (advance (p, v) < 0)
			 done = TRUE;
	       }

	       /* check to see if it's time to stop rendering */
	       if (done)
		    v->active = FALSE;

	       /* decrease the voices offset if applicable */
	       if (v->offset > 0)
	       {
		    if (v->offset > nframes)
		    {
			 v->offset -= nframes;
		    }
		    else
		    {
			 v->offset = 0;
		    }
	       }

	       /* overflows bad, OVERFLOWS BAD! */
	       if (v->active && (v->posi < 0 || v->posi >= p->sample.frames))
	       {
		    debug ("overflow! NO! BAD CODE! DIE DIE DIE!\n");
		    debug ("v->posi == %d, p->sample.frames == %d\n", v->posi, p->sample.frames);
		    v->active = 0;
	       }
	  }
     }
}

/* deactivate all active patches matching given criteria */
void patch_release (int chan, int note, int offset)
{
     int i;

     for (i = 0; i < PATCH_COUNT; i++)
     {
	  if (patches[i].active
	      && !(patches[i].play_mode & PATCH_PLAY_SINGLESHOT)
	      && patches[i].channel == chan
	      && (patches[i].note == note
		  || (patches[i].range > 0
		      && note >= patches[i].lower_note
		      && note <= patches[i].upper_note)))
	  {

	       patch_release_patch (&patches[i], note, offset);
	  }
     }

     return;
}

/* deactivate a single patch with a given id */
void patch_release_with_id (int id, int note, int offset)
{
     if (id < 0 || id >= PATCH_COUNT)
	  return;
     if (!patches[id].active)
	  return;
     if (patches[id].play_mode & PATCH_PLAY_SINGLESHOT)
	  return;

     patch_release_patch (&patches[id], note, offset);
     return;
}

/* render nframes of all active patches into buf */
void patch_render (float *buf, int nframes)
{
     int i;

     for (i = 0; i < PATCH_COUNT; i++)
     {
	  if (patches[i].active)
	  {
	       if (pthread_mutex_trylock (&patches[i].sample.mutex) != 0)
	       {
		    continue;
	       }
	       if (patches[i].sample.sp == NULL)
	       {
		    pthread_mutex_unlock (&patches[i].sample.mutex);
		    continue;
	       }
	       patch_render_patch (&patches[i], buf, nframes);
	       pthread_mutex_unlock (&patches[i].sample.mutex);
	  }
     }
}

/* triggers all patches matching criteria */
void patch_trigger (int chan, int note, float vel, Tick ticks,
		    int offset)
{
     static int idp[PATCH_COUNT];	/* holds all patches to be activated */
     int i, j;

     /* We gather up all of the patches that need to be activated here
      * so that we can run their cuts and then trigger them without
      * having to find them twice.  We have to make sure that we do
      * cutting before triggering, otherwise patches which listen on
      * the same note and have the same cut/cut_by values will end up
      * stepping over each other before they both are heard.
      */

     for (i = j = 0; i < PATCH_COUNT; i++)
     {
	  if (patches[i].active
	      && patches[i].channel == chan
	      && (patches[i].note == note
		  || (patches[i].range
		      && note >= patches[i].lower_note
		      && note <= patches[i].upper_note)))
	  {

	       idp[j++] = i;
	  }
     }

     /* do cuts */
     for (i = 0; i < j; i++)
	  patch_cut_patch (&patches[idp[i]], offset);

     /* do triggers */
     for (i = 0; i < j; i++)
	  patch_trigger_patch (&patches[idp[i]], note, vel, ticks, offset);
}

/* activate a single patch with given id */
void patch_trigger_with_id (int id, int note, float vel, Tick ticks,
			    int offset)
{
     if (id < 0 || id >= PATCH_COUNT)
	  return;
     if (!patches[id].active)
	  return;

     patch_cut_patch (&patches[id], offset);
     patch_trigger_patch (&patches[id], note, vel, ticks, offset);
     return;
}

/*****************************************************************************/
/*************************** ENVELOPE SETTERS ********************************/
/*****************************************************************************/

/* sets how much the envelope affects a param */
int patch_set_env_amount (int id, PatchParamType param, float amt)
{
     PatchParam *p;
     int err;

     if (param == PATCH_PARAM_VOLUME)
	  return PATCH_PARAM_INVALID;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     if (amt < -1.0 || amt > 1.0)
	  return PATCH_PARAM_INVALID;

     p->env_amt = amt;
     return 0;
}

/* sets the attack length in seconds */
int patch_set_env_attack (int id, PatchParamType param, float secs)
{
     PatchParam *p;
     int err;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     if (secs < 0.0)
	  return PATCH_PARAM_INVALID;

     p->env.attack = secs;
     return 0;
}

/* sets the decay length in seconds */
int patch_set_env_decay (int id, PatchParamType param, float secs)
{
     PatchParam *p;
     int err;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     if (secs < 0.0)
	  return PATCH_PARAM_INVALID;

     p->env.decay = secs;
     return 0;
}

/* sets the sustain level */
int patch_set_env_sustain (int id, PatchParamType param, float level)
{
     PatchParam *p;
     int err;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     if (level < 0.0 || level > 1.0)
	  return PATCH_PARAM_INVALID;

     p->env.sustain = level;
     return 0;
}

/* sets the release length in seconds */
int patch_set_env_release (int id, PatchParamType param, float secs)
{
     PatchParam *p;
     int err;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     if (secs < 0.0)
	  return PATCH_PARAM_INVALID;

     if (secs < PATCH_MIN_RELEASE)
	  secs = PATCH_MIN_RELEASE;

     p->env.release = secs;
     return 0;
}

/*****************************************************************************/
/*************************** ENVELOPE GETTERS ********************************/
/*****************************************************************************/

/* places the amount the envelope is used into val */
int patch_get_env_amount (int id, PatchParamType param, float *val)
{
     PatchParam *p;
     int err;

     if (param == PATCH_PARAM_VOLUME)
	  return PATCH_PARAM_INVALID;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     *val = p->env_amt;
     return 0;
}

/* places the attack length in seconds into val */
int patch_get_env_attack (int id, PatchParamType param, float *val)
{
     PatchParam *p;
     int err;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     *val = p->env.attack;
     return 0;
}

/* places the decay length in seconds into val */
int patch_get_env_decay (int id, PatchParamType param, float *val)
{
     PatchParam *p;
     int err;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     *val = p->env.decay;
     return 0;
}

/* places the sustain level into val */
int patch_get_env_sustain (int id, PatchParamType param, float *val)
{
     PatchParam *p;
     int err;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     *val = p->env.sustain;
     return 0;
}

/* places the release length in seconds into val */
int patch_get_env_release (int id, PatchParamType param, float *val)
{
     PatchParam *p;
     int err;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     *val = p->env.release;

     /* we hide the fact that we have a min release value from the
      * outside world (they're on a need-to-know basis) */
     if (*val <= PATCH_MIN_RELEASE)
	  *val = 0;
     return 0;
}

/************************************************************************/
/*************************** LFO SETTERS ********************************/
/************************************************************************/

/* set which lfo this patch should listen, or negative if none */
int patch_set_lfo (int id, PatchParamType param, int lfo_id)
{
     PatchParam *p;
     int err;

     if (!isok (id))
	  return PATCH_ID_INVALID;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     if (lfo_id >= LFO_COUNT)
	  return PATCH_PARAM_INVALID;

     p->lfo = lfo_id;
     return 0;
}

/* set the amount of the lfo to add to param */
int patch_set_lfo_amount (int id, PatchParamType param, float amt)
{
     PatchParam *p;
     int err;

     if (!isok (id))
	  return PATCH_ID_INVALID;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     if (amt < -1.0 || amt > 1.0)
	  return PATCH_PARAM_INVALID;

     p->lfo_amt = amt;
     return 0;
}

/* get the id of the lfo used, or negative if none */
int patch_get_lfo (int id, PatchParamType param, int *val)
{
     PatchParam *p;
     int err;

     if (!isok (id))
	  return PATCH_ID_INVALID;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     *val = p->lfo;
     return 0;
}

/* get the amount of the lfo that is added to param */
int patch_get_lfo_amount (int id, PatchParamType param, float *val)
{
     PatchParam *p;
     int err;

     if (!isok (id))
	  return PATCH_ID_INVALID;

     if ((err = patch_get_param (&p, id, param)) < 0)
	  return err;

     *val = p->lfo_amt;
     return 0;
}

/*****************************************************************************/
/*************************** PARAMETER SETTERS *******************************/
/*****************************************************************************/

/* sets channel patch listens on */
int patch_set_channel (int id, int channel)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (channel < 0 || channel > 15)
	  return PATCH_CHANNEL_INVALID;

     patches[id].channel = channel;
     return 0;
}

/* sets the cut signal this patch emits when activated */
int patch_set_cut (int id, int cut)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     patches[id].cut = cut;
     return 0;
}

/* sets the cut signal that terminates this patch if active */
int patch_set_cut_by (int id, int cut_by)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     patches[id].cut_by = cut_by;
     return 0;
}

/* sets filter cutoff frequency */
int patch_set_cutoff (int id, float freq)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (freq < 0.0 || freq > 1.0)
	  return PATCH_PARAM_INVALID;

     patches[id].ffreq.val = freq;
     return 0;
}

/* sets the start loop point */
int patch_set_loop_start (int id, int start)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (patches[id].sample.sp == NULL)
	  return 0;

     if (start < patches[id].sample_start)
	  start = patches[id].sample_start;
     else if (start > patches[id].sample_stop)
	  start = patches[id].sample_stop;

     patches[id].loop_start = start;
     if (start > patches[id].loop_stop)
	  patches[id].loop_stop = start;

     return 0;
}

/* sets the stopping loop point */
int patch_set_loop_stop (int id, int stop)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (patches[id].sample.sp == NULL)
	  return 0;

     if (stop > patches[id].sample_stop)
	  stop = patches[id].sample_stop;
     else if (stop < patches[id].sample_start)
	  stop = patches[id].sample_start;

     patches[id].loop_stop = stop;
     if (stop < patches[id].loop_start)
	  patches[id].loop_start = stop;

     return 0;
}

/* sets the lower note of a patch's range */
int patch_set_lower_note (int id, int note)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (note < 0 || note > 127)
	  return PATCH_NOTE_INVALID;

     patches[id].lower_note = note;
     return 0;
}

/* sets the name */
int patch_set_name (int id, const char *name)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     strncpy (patches[id].name, name, PATCH_MAX_NAME);
     return 0;
}

/* sets the root note */
int patch_set_note (int id, int note)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (note < 0 || note > 127)
	  return PATCH_NOTE_INVALID;

     patches[id].note = note;
     return 0;
}

/* sets the panorama */
int patch_set_panning (int id, float pan)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (pan < -1.0 || pan > 1.0)
	  return PATCH_PAN_INVALID;

     patches[id].pan.val = pan;
     return 0;
}

/* sets the play mode */
int patch_set_play_mode (int id, PatchPlayMode mode)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     /* verify direction */
     if (mode & PATCH_PLAY_FORWARD)
     {
	  if (mode & PATCH_PLAY_REVERSE)
	  {
	       return PATCH_PLAY_MODE_INVALID;
	  }
     }
     else if (mode & PATCH_PLAY_REVERSE)
     {
	  if (mode & PATCH_PLAY_FORWARD)
	  {
	       return PATCH_PLAY_MODE_INVALID;
	  }
     }
     else
     {
	  return PATCH_PLAY_MODE_INVALID;
     }

     /* verify duration */
     if (mode & PATCH_PLAY_SINGLESHOT)
     {
	  if ((mode & PATCH_PLAY_TRIM) || (mode & PATCH_PLAY_LOOP))
	  {
	       return PATCH_PLAY_MODE_INVALID;
	  }
     }
     else if (mode & PATCH_PLAY_TRIM)
     {
	  if ((mode & PATCH_PLAY_SINGLESHOT) || (mode & PATCH_PLAY_LOOP))
	  {
	       return PATCH_PLAY_MODE_INVALID;
	  }
     }
     else if (mode & PATCH_PLAY_LOOP)
     {
	  if ((mode & PATCH_PLAY_SINGLESHOT) || (mode & PATCH_PLAY_TRIM))
	  {
	       return PATCH_PLAY_MODE_INVALID;
	  }
     }

     /* make sure pingpong isn't frivolously set (just for style
      * points) */
     if ((mode & PATCH_PLAY_PINGPONG) && !(mode && PATCH_PLAY_LOOP))
     {
	  return PATCH_PLAY_MODE_INVALID;
     }

     patches[id].play_mode = mode;
     return 0;
}

/* set patch to listen to a range of notes if non-zero */
int patch_set_range (int id, int range)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     patches[id].range = range;
     return 0;
}

/* set the filter's resonance */
int patch_set_resonance (int id, float reso)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (reso < 0.0 || reso > 1.0)
	  return PATCH_PARAM_INVALID;

     patches[id].freso.val = reso;
     return 0;
}

/* set the point within the sample to begin playing */
int patch_set_sample_start (int id, int start)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (patches[id].sample.sp == NULL)
	  return 0;

     if (start < 0)
     {
	  debug ("refusing to set negative sample start point\n");
	  return PATCH_PARAM_INVALID;
     }

     if (start > patches[id].sample_stop)
     {
	  debug ("refusing to set incongruous sample start point\n");
	  return PATCH_PARAM_INVALID;
     }

     patches[id].sample_start = start;
     return 0;
}

/* set the point within the sample to stop playing */
int patch_set_sample_stop (int id, int stop)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (patches[id].sample.sp == NULL)
	  return 0;

     if (stop >= patches[id].sample.frames)
     {
	  debug
	       ("refusing to set sample stop greater than sample frame count\n");
	  return PATCH_PARAM_INVALID;
     }

     if (stop < patches[id].sample_start)
     {
	  debug ("refusing to set incongruous sample stop point\n");
	  return PATCH_PARAM_INVALID;
     }

     patches[id].sample_stop = stop;
     return 0;
}

/* set the upper note of a patch's range */
int patch_set_upper_note (int id, int note)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (note < 0 || note > 127)
	  return PATCH_NOTE_INVALID;

     patches[id].upper_note = note;
     return 0;
}

/* set the volume */
int patch_set_volume (int id, float vol)
{

     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (vol < 0 || vol > 1.0)
	  return PATCH_VOL_INVALID;

     patches[id].vol.val = vol;
     return 0;
}

/*****************************************************************************/
/*************************** PARAMETER GETTERS********************************/
/*****************************************************************************/

/* get the channel the patch listens on */
int patch_get_channel (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].channel;
}

/* get the cut signal */
int patch_get_cut (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].cut;
}

/* get the cut-by signal */
int patch_get_cut_by (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].cut_by;
}

/* get the filter cutoff value */
float patch_get_cutoff (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].ffreq.val;
}

/* get the display index */
int patch_get_display_index (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].display_index;
}

/* get the number of frame in the sample */
int patch_get_frames (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;

     if (patches[id].sample.sp == NULL)
	  return 0;

     return patches[id].sample.frames;
}

/* get the starting loop point */
int patch_get_loop_start (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].loop_start;
}

/* get the stopping loop point */
int patch_get_loop_stop (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].loop_stop;
}

/* get the lower note */
int patch_get_lower_note (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].lower_note;
}

/* get the name */
char *patch_get_name (int id)
{
     char *name;

     if (id < 0 || id >= PATCH_COUNT || !patches[id].active)
	  name = strdup ("\0");
     else
	  name = strdup (patches[id].name);

     return name;
}

/* get the root note */
int patch_get_note (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].note;
}

/* get the panorama */
float patch_get_panning (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].pan.val;
}

/* get the play mode */
PatchPlayMode patch_get_play_mode (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].play_mode;
}

/* get whether a range of notes is used or not */
int patch_get_range (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].range;
}

/* get the filter's resonance amount */
float patch_get_resonance (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].freso.val;
}

/* get a pointer to the sample data */
float const *patch_get_sample (int id)
{
     if (id < 0 || id >= PATCH_COUNT || !patches[id].active)
	  return NULL;

     return patches[id].sample.sp;
}

/* get the name of the sample file */
char *patch_get_sample_name (int id)
{
     char *name;

     if (id < 0 || id >= PATCH_COUNT || !patches[id].active)
	  name = strdup ("\0");
     else
	  name = strdup (patches[id].sample.name);
     return name;
}

/* get the starting playback point */
int patch_get_sample_start (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].sample_start;
}

/* get the stopping playback point */
int patch_get_sample_stop (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].sample_stop;
}

/* get the upper note */
int patch_get_upper_note (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].upper_note;
}

/* get the volume */
float patch_get_volume (int id)
{
     if (!isok (id))
	  return PATCH_ID_INVALID;
     return patches[id].vol.val;
}
